#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <openssl/md5.h>
#include <errno.h>
#include <limits.h>

#define BLOCKSIZE 4096

int hash_md5_symlink(char *filename,
                     int flags,
                     unsigned char *hash)
{
  char buf[PATH_MAX];
  int n;
  
  n = readlink(filename, buf, sizeof(buf));
  if (n < 0) {
    return -1;
  } else if (n == PATH_MAX) {
    errno = ENAMETOOLONG;
    return -1;
  } else {
    MD5_CTX ctx;

    MD5_Init(&ctx);
    MD5_Update(&ctx, buf, n);
    MD5_Final(hash, &ctx);
  }
  return 0;
}

int hash_md5_file(char *filename,
                  int flags,
                  unsigned char *hash)
{
  int fd;
  
  fd = open (filename, O_RDONLY);
  if (fd < 0) {
    return -1;
  } else {
    MD5_CTX ctx;
    unsigned char buffer[BLOCKSIZE];
    int count;
    time_t now;
      
    time(&now);
    MD5_Init(&ctx);
    while ((count = read(fd, buffer, BLOCKSIZE)) > 0) {
      MD5_Update(&ctx, buffer, count);
    }
    MD5_Final(hash, &ctx);

    goto out_close;
  }
out_close:
  close(fd);
#if 0
  char digest[128];
  int length = 0, recalc = 1;
  char md5[33];

  if (options->flags & FLAGS_CLEAR_XATTR) {
    if (removexattr(filename, XATTR_MD5SUM) != 0) {
      if (errno != ENOATTR) {
        fprintf(stderr, "Failed to remove xattr '%s'\n", filename);
      }
    } else if (options->flags & FLAGS_VERBOSE) {
      fprintf(stderr, "Removing xattr for '%s'\n", filename);     
    }
    goto out;
  }
  if (options->flags & FLAGS_READ_XATTR) {
    length = getxattr(filename, XATTR_MD5SUM, digest, 128);
  }
  if (length > 0) {
#define FORMAT "t=%" PRIu64 ",mtime=%" PRIu64 ".%" PRIu32 ",md5=%32s"
    uint64_t t;
    struct {
      uint64_t tv_sec;
      uint32_t tv_nsec;
    } mtime;
    int items = sscanf(digest, FORMAT, &t, &mtime.tv_sec, &mtime.tv_nsec, md5);
    if (items == 4 &&
        ((options->flags & FLAGS_MAX_AGE) == 0 || t >= options->max_age) &&
        mtime.tv_sec == buf.st_mtim.tv_sec &&
        mtime.tv_nsec == buf.st_mtim.tv_nsec) {
      recalc = 0;
   } else {
      if (removexattr(filename, XATTR_MD5SUM) != 0) {
        if (errno != ENOATTR) {
          fprintf(stderr, "Failed to remove xattr '%s' (mtime)\n", filename);
        }
      } else if (options->flags & FLAGS_VERBOSE) {
        fprintf(stderr, "Removing xattr for '%s'\n", filename);   
      }
    }
  }
  if (recalc) {
    FILE *f;
    unsigned char sum[16];

    f = fopen (filename, "r");
    if (!f) {
      sprintf(md5, "_unreadable_");
      if (options->flags & (FLAGS_READ_XATTR|FLAGS_WRITE_XATTR)) {
        /* Only remove md5sum for unreadble file if we are root */
        if (removexattr(filename, XATTR_MD5SUM) != 0) {
          if (errno != ENOATTR) {
            fprintf(stderr, "Failed to remove xattr '%s' (unreadable)\n", 
                    filename);
          }
        } else if (options->flags & FLAGS_VERBOSE) {
          fprintf(stderr, "Removing xattr for '%s'\n", filename);         
        }
      }
    } else {
      int i;
      time_t now;
      
      time(&now);
      
      md5_stream(f, &sum);
      for (i = 0 ; i < 16 ; i++) {
        sprintf(&md5[i*2], "%2.2x", sum[i]);
      }
      fclose(f);
      if (options->flags & FLAGS_WRITE_XATTR) {
        int len = snprintf(digest, sizeof(digest),
                           "t=%" PRIu64 ",mtime=%" PRIu64 ".%9.9u,md5=%32s",
                           (uint64_t)now,  
                           (uint64_t)buf.st_mtim.tv_sec,
                           (uint32_t) buf.st_mtim.tv_nsec, 
                           md5);
        if (setxattr(filename, XATTR_MD5SUM, digest, len, 0) != 0) {
          fprintf(stderr, "Failed to set xattr '%s'\n", filename);
        } else if (options->flags & FLAGS_VERBOSE) {
          fprintf(stderr, "Setting xattr for '%s'\n", filename);          
        }
      }
    }
  }
  printentry(buf.st_size, buf.st_mtime, 
             buf.st_uid, buf.st_gid, buf.st_mode,
             md5, 'F', filename);
out:
  ;
}

void dofile(char *filename, dev_t *dev, struct options *options) 
{
  struct stat64 buf;
  int err;
  
  if (getuid() != 0 &&
      faccessat(AT_FDCWD, filename, R_OK, AT_SYMLINK_NOFOLLOW) != 0) {
    /* Bail out if no access */
    fprintf(stderr, "Failed to access '%s'\n", filename);
    goto out;
  }
  err = lstat64(filename, &buf);
  if (err == 0 && dev == 0) { dev = &buf.st_dev; }
  if (err != 0) {
    fprintf(stderr, "Failed to lstat '%s'\n", filename);
  } else if (*dev != buf.st_dev) {
    fprintf(stderr, "skipping mountpoint '%s'\n", filename);
  } else if (filenameislegal(filename)) {
    if (S_ISREG(buf.st_mode)) {
      md5sum_file(filename, buf, options);
    } else if (S_ISDIR(buf.st_mode)) {
      if (! (options->flags & FLAGS_CLEAR_XATTR)) {
        printentry(0, 0, buf.st_uid, buf.st_gid, buf.st_mode,
                   "", 'D', filename);
      }
      if (filename[0] == 0) { filename = "."; }
      dodir(filename, dev, options);
    } else if (options->flags & FLAGS_CLEAR_XATTR) {
      goto out;
    } else if (S_ISLNK(buf.st_mode)) {
      char *link = malloc(buf.st_size + 1);
      err = readlink(filename, link, buf.st_size);
      if (err != buf.st_size) {
        fprintf(stderr, "failed to readlink: %s\n", filename);
      } else {
        char md5[33];
        unsigned char sum[16];
        int i;

        md5_buffer(link, buf.st_size, &sum);
        for (i = 0 ; i < 16 ; i++) {
          sprintf(&md5[i*2], "%2.2x", sum[i]);
        }
        printentry(buf.st_size, buf.st_mtime,  
                   buf.st_uid, buf.st_gid, buf.st_mode,
                   md5, 'L', filename);
      }
      free(link);
    } else {
      char kind;

      switch(buf.st_mode & S_IFMT) {
        case S_IFSOCK: { kind = 'S'; } break;
        case S_IFBLK: { kind = 'B'; } break;
        case S_IFCHR: { kind = 'C'; } break; 
        case S_IFIFO: { kind = 'F'; } break;
        default: { kind = 'U'; } 
      }
      printentry(buf.st_size, buf.st_mtime, 
                 buf.st_uid, buf.st_gid, buf.st_mode,
                 "", kind, filename);
    }
  }
out:
  ;
#endif
    
/*
  if (options->flags & FLAGS_READ_XATTR) {
    length = getxattr(filename, XATTR_MD5SUM, digest, 128);
  }
int MD5_Init(MD5_CTX *c);
        int MD5_Update(MD5_CTX *c, const void *data,
                         unsigned long len);
        int MD5_Final(unsigned char *md, MD5_CTX *c);
char data[] = "data to hash";
char hash[SHA512_DIGEST_LENGTH];
SHA512(data, sizeof(data) - 1, hash);
*/
  return 0;
}
